'''
Allows sioyek to automatically download papers from scihub by clicking on their names
Requires the following python packages to be installed:
* PyPaperBot: https://github.com/ferru97/PyPaperBot
* habanero: https://github.com/sckott/habanero

Here is an example `prefs_user.config` file which uses this script:

    execute_command_l python /path/to/paper_downloader.py "%{sioyek_path}" "%{paper_name}"
    control_click_command execute_command_l

Now, you can control+click on paper names to download them and open them in sioyek.
'''

# where to put downloaded papers, if it is None, we use a papers folder next to this script
PAPERS_FOLDER_PATH = None
SIOYEK_PATH = None
PYTHON_EXECUTABLE = 'python'

import os
import pathlib
import subprocess
import sys
import json
import time
import regex
import fitz

from habanero import Crossref

def log(file_name, text):
    with open(file_name, 'a') as f:
        f.write(str(text) + '\n')

def show_status(string):
    if len(string) > 0:
        subprocess.run([SIOYEK_PATH, '--execute-command', 'set_status_string','--execute-command-data', string])
    else:
        subprocess.run([SIOYEK_PATH, '--execute-command', 'clear_status_string'])
        

def clean_paper_name(paper_name):
    first_quote_index = -1
    last_quote_index = -1

    try:
        first_quote_index = paper_name.index('"')
        last_quote_index = paper_name.rindex('"')
    except ValueError as e:
        pass

    if first_quote_index != -1 and last_quote_index != -1 and (last_quote_index - first_quote_index) > 10:
        return paper_name[first_quote_index + 1:last_quote_index]

    paper_name = paper_name.strip()
    if paper_name.endswith('.'):
        paper_name = paper_name[:-1]

    return paper_name

def is_file_a_paper_with_name(file_name, paper_name):
    doc = fitz.open(file_name)
    page_text = doc.get_page_text(0)
    doc.close()
    if regex.search('(' + regex.escape(paper_name) + '){e<=6}', page_text, flags=regex.IGNORECASE):
        return True
    return False


class ListingDiff():
    def __init__(self, path):
        self.path = path

    def get_listing(self):
        return set(os.listdir(self.path))

    def __enter__(self):
        self.listing = self.get_listing()
        return self
    
    def new_files(self):
        return list(self.get_listing().difference(self.listing))
    
    def new_pdf_files(self):
        return [f for f in self.new_files() if f.endswith('.pdf')]
    
    def reset(self):
        self.listing = self.get_listing()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        pass

def get_cached_doi_map():
    json_path = get_papers_folder_path() / 'doi_map.json'
    if os.path.exists(json_path):
        with open(json_path, 'r') as infile:
            return json.load(infile)
    else:
        return dict()

def write_cache_doi_map(doi_map):
    json_path = get_papers_folder_path() / 'doi_map.json'
    with open(json_path, 'w') as outfile:
        json.dump(doi_map, outfile)


def get_papers_folder_path_():
    if PAPERS_FOLDER_PATH != None:
        return pathlib.Path(PAPERS_FOLDER_PATH)
    return pathlib.Path(os.path.dirname(os.path.realpath(__file__))) / 'papers'

def get_papers_folder_path():
    path = get_papers_folder_path_()
    path.mkdir(parents=True, exist_ok=True)
    return path

def get_doi_with_name(paper_name):
    crossref = Crossref()
    response = crossref.works(query=paper_name)
    if len(response['message']['items']) == 0:
        return None
    return response['message']['items'][0]['DOI']

def download_paper_with_doi(doi_string, paper_name, doi_map):
    download_dir = get_papers_folder_path()
    methods_to_try = [
        ('scihub', [PYTHON_EXECUTABLE, '-m', 'PyPaperBot', '--doi={}'.format(doi_string), '--dwn-dir={}'.format(download_dir)]),
        ('google scholar', [PYTHON_EXECUTABLE, '-m', 'PyPaperBot', '--query={}'.format(paper_name), '--scholar-pages=1', '--scholar-results=1', '--dwn-dir={}'.format(download_dir)])
    ]
    with ListingDiff(download_dir) as listing_diff:

        ignored_files = []

        for method_name, method_args in methods_to_try:
            show_status('trying to download "{}" from {}'.format(paper_name, method_name))
            try:
                subprocess.run(method_args)
            except Exception as e:
                show_status('error in download from {}'.format(method_name))
            pdf_files = listing_diff.new_pdf_files()
            if len(pdf_files) > 0:
                returned_file = download_dir / pdf_files[0]
                if is_file_a_paper_with_name(str(returned_file), paper_name):
                    doi_map[doi_string] = str(returned_file)
                    write_cache_doi_map(doi_map)
                    return returned_file
                else:
                    ignored_files.append(returned_file)
                    listing_diff.reset()

        if len(ignored_files) > 0:
            show_status('could not find a suitable paper, this is a throw in the dark')
            return ignored_files[0]

    return None

def get_paper_file_name_with_doi_and_name(doi, paper_name):

    doi_map = get_cached_doi_map()
    if doi in doi_map.keys():
        if os.path.exists(doi_map[doi]):
            return doi_map[doi]
    return download_paper_with_doi(doi, paper_name, doi_map)


if __name__ == '__main__':

    SIOYEK_PATH = sys.argv[1]
    paper_name = clean_paper_name(sys.argv[2])

    show_status('finding doi ...')
    try:
        doi = get_doi_with_name(sys.argv[2])
        if doi:
            # show_status('downloading doi: {}'.format(doi))
            file_name = get_paper_file_name_with_doi_and_name(doi, paper_name)
            show_status("")
            if file_name:
                subprocess.run([SIOYEK_PATH, str(file_name), '--new-window'])
        else:
            show_status('doi not found')
            time.sleep(5)
            show_status("")
    except Exception as e:
        show_status('error: {}'.format(str(e)))
        time.sleep(5)
        show_status("")


'''
example `prefs_user.config`:
    # start reading from highlighted line
    execute_command_a python D:\labs2\tts\second-try\server_read.py "%1" %4 "%6"
    # stop reading
    execute_command_b python D:\labs2\tts\second-try\server_stop.py
    # keep highlighting the current line being read
    execute_command_c python D:\labs2\tts\second-try\server_follow.py
    # stop highlighting the current line being read
    execute_command_d python D:\labs2\tts\second-try\server_unfollow.py
    # start the tts server (should be running before executing previous commands)
    execute_command_e python D:\labs2\tts\second-try\manager_server.py

example `keys_user.config`:
    execute_command_a r
    execute_command_b <S-r>
    execute_command_e <C-<f5>>
    # when we manually move the line, stop following it
    move_visual_mark_down;execute_command_d j
    move_visual_mark_up;execute_command_d k
'''

from flask import Flask
from flask import request
import threading
import time
import datetime
import subprocess

import hashlib
import json
import os

import fitz
import pygame
from pygame import mixer
import platform

if (platform.system() == "Darwin"):
    # cached audio files will be stored in this directory
    DATA_FOLDER = "./tts-output/"
    SIOYEK_PATH = "sioyek"
    POWER_SHELL = "pwsh"
else:
    # cached audio files will be stored in this directory
    DATA_FOLDER = "D:\\tts-output\\"
    SIOYEK_PATH = "sioyek.exe"
    POWER_SHELL = "powershell"

if(not os.path.exists(DATA_FOLDER)):
    os.mkdir(DATA_FOLDER)

def lcs(pattern_1, pattern_2):
    m = len(pattern_1)
    n = len(pattern_2)
    # dp will store solutions as the iteration goes on
    dp = [
        [None] * (n + 1) for item in range(m + 1)
    ]
    for i in range(m + 1):
        for j in range(n + 1):
            if i == 0 or j == 0:
                dp[i][j] = 0
            elif pattern_1[i - 1] == pattern_2[j - 1]:
                dp[i][j] = dp[i - 1][j - 1] + 1
            else :
                dp[i][j] = max(dp[i - 1][j], dp[i][j - 1])
    return dp[m][n]

def convert_text_to_appropriate_format_for_tts(text):
    lines = text.split('\n')
    n_lines = len(lines)
    out_lines = []
    lines_to_merge = []
    
    for i, line in enumerate(lines):
        if i < (n_lines-1) and line[-1] == '-':
            lines_to_merge.append(line[:-1])
        elif i < (n_lines-1) and line[-1] != '.':
            lines_to_merge.append(line + ' ')
        else:
            lines_to_merge.append(line)
            line_text_raw = ''.join(lines_to_merge)
            line_text_clean = ''.join(x for x in line_text_raw if ord(x) < 127)
            out_lines.append(line_text_clean)
            lines_to_merge = []
    return ('\n'.join(out_lines)).replace('"', '""')


def parse_alignment_file(filename):
    with open(filename, 'r') as infile:
        data = json.load(infile)
    output = []
    for i, frag in enumerate(data['fragments']):
        output.append((float(frag['begin']), float(frag['end']), frag['lines'][0], i))
    return output

class Manager:

    def __init__(self):
        self.should_follow_lines = True
        self.doc_cache = dict()
        self.alignment_cache = dict()
        self.worker_threads = []
        self.background_mutex = threading.Lock()
        self.current_track = None
        self.offset = None
        self.current_state = None
        self.last_clean_time = datetime.datetime.now()
        mixer.init()
    
    def get_doc(self, path):
        if path not in self.doc_cache.keys():
            self.doc_cache[path] = fitz.open(path)
        return self.doc_cache[path]
    
    def get_page_text(self, path, page):
        return self.get_doc(path).get_page_text(page)

    def get_page_text_tts(self, path, page):
        return convert_text_to_appropriate_format_for_tts(self.get_page_text(path, page))

    def get_digest(self, path, page, fast):
        text = self.get_page_text(path, page)
        digest = hashlib.sha1(text.encode('utf-8')).hexdigest()
        if fast:
            return digest + "_fast"
        else:
            return digest + "_good"

    def get_alignment(self, digest):
        file_name = DATA_FOLDER + digest + "_alignment.json"
        if digest not in self.alignment_cache.keys():
            self.alignment_cache[digest] = parse_alignment_file(file_name)
        return self.alignment_cache[digest]
    
    def has_fast_tts(self, path, page):
        digest = self.get_digest(path, page, True)
        return os.path.exists(DATA_FOLDER + digest + "_alignment.json")

    def has_good_tts(self, path, page):
        digest = self.get_digest(path, page, False)
        return os.path.exists(DATA_FOLDER + digest + "_alignment.json")

    def enable_line_follow(self):
        self.should_follow_lines = True

    def disable_line_follow(self):
        self.should_follow_lines = False
    
    def create_files_or_wait_for_background(self, path, page_number, fast):
        with self.background_mutex:
            job = self.find_background_job(path, page_number, fast)
            if job != None:
                job.join()
            else:
                self.create_files(path, page_number, fast)


    def create_files(self, path, page_number, fast):
            digest = self.get_digest(path, page_number, fast)


            tts_text_file_name = DATA_FOLDER + digest + "_tts.txt"
            align_text_file_name = DATA_FOLDER + digest + "_align.text"
            align_output_file_name = DATA_FOLDER + digest + "_alignment.json"
            audio_file_name = DATA_FOLDER + digest + ".wave"
            ogg_file_name = DATA_FOLDER + digest + ".ogg"

            tts_text = self.get_page_text_tts(path, page_number)
            align_text = self.get_page_text(path, page_number)

            with open(tts_text_file_name, "w", encoding='utf8') as outfile:
                outfile.write(tts_text)

            with open(align_text_file_name, "w", encoding='utf8') as outfile:
                outfile.write(align_text)

            if fast:
                subprocess.run([POWER_SHELL, os.path.dirname(__file__) + "\\generator.ps1", tts_text_file_name, audio_file_name])
            else:
                subprocess.run([POWER_SHELL, os.path.dirname(__file__) + "\\generator2.ps1", tts_text_file_name, audio_file_name])

            subprocess.run(["sox", audio_file_name, ogg_file_name, 'tempo', '1.5'])
            subprocess.run([POWER_SHELL, os.path.dirname(__file__) + "\\aligner.ps1", ogg_file_name, align_text_file_name, align_output_file_name])

    def find_background_job(self, path, page_number, fast):
        for thread_path, thread_page, thread_fast, thread in self.worker_threads:
            if (thread_path, thread_page, thread_fast) == (path, page_number, fast):
                return thread
        return None

    def add_background_job(self, path, page_number, fast):
        with self.background_mutex:
            if self.find_background_job(path, page_number, fast) != None:
                return False
            thread = threading.Thread(target=self.create_files, args=(path, page_number, fast))
            self.worker_threads.append((path, page_number, fast, thread))
            thread.start()
            return True
            

    def get_best_line_index(self, alignment, line_text):
        best_line_index = -1
        best_line_score = -1

        for i, line_data in enumerate(alignment):
            current_line_text = line_data[2]
            score = lcs(current_line_text, line_text)
            if score > best_line_score:
                best_line_score = score
                best_line_index = i
        return best_line_index

    def set_track_with_digest(self, digest):
        file_path = DATA_FOLDER + digest + ".ogg"
        if self.current_track != file_path:
            mixer.music.stop()
            mixer.music.unload()
            mixer.music.load(file_path)
            mixer.music.play()
            mixer.music.set_volume(0.7)
            self.current_track = file_path

    def goto_line_with_text(self, path, page_number, line_text):
        digest = None

        if self.has_good_tts(path, page_number):
            digest = self.get_digest(path, page_number, False)
        elif self.has_fast_tts(path, page_number):
            digest = self.get_digest(path, page_number, True)
        else:
            self.create_files_or_wait_for_background(path, page_number, True)
            self.add_background_job(path, page_number, False)
            digest = self.get_digest(path, page_number, True)
        
        alignment = self.get_alignment(digest)
        if line_text != None:
            line_index = self.get_best_line_index(alignment, line_text)
        else:
            line_index = 0

        line_begin = alignment[line_index][0]
        self.set_track_with_digest(digest)
        mixer.music.play()
        mixer.music.set_pos(line_begin) #!!!!!!!!!!!!!!!!! ogg processing is accurate no need for hack mp3
        self.offset = line_begin
        self.current_state = (path, page_number, line_index, digest)
        self.enable_line_follow()

    def find_pos_index(self, pos):
        alignments = self.get_alignment(self.current_state[3])

        for i, (begin, end, _, _) in enumerate(alignments):
            if pos >= begin and pos <= end:
                return i

        return -1

    def highlight_line(self, line_text):
        subprocess.run([SIOYEK_PATH, "--focus-text", '"{}"'.format(line_text), "--focus-text-page", str(self.current_state[1]), "--nofocus"])

    def better_audio_is_available(self):
        if self.current_state[-1].endswith("_fast") and self.has_good_tts(self.current_state[0], self.current_state[1]):
            return True
        return False

    def goto_line_if_changed(self):
        current_pos = mixer.music.get_pos()
        if current_pos == None:
            return
        if current_pos != -1:
            if self.offset == None:
                pos = current_pos / 1000
            else:
                pos = current_pos / 1000 + self.offset
            index = self.find_pos_index(pos)
            if index != self.current_state[2]:
                new_line_text = self.get_alignment(self.current_state[3])[index][2]
                if self.better_audio_is_available():
                    self.goto_line_with_text(self.current_state[0], self.current_state[1], new_line_text)
                else:
                    self.highlight_line(new_line_text)
                    self.current_state = self.current_state[0], self.current_state[1], index, self.current_state[3]


    def clean_old_files(self):
        if self.last_clean_time == None or (datetime.datetime.now() - self.last_clean_time).seconds > 1000:
            # find old files
            self.last_clean_time = datetime.datetime.now()

            for file in os.listdir(DATA_FOLDER):
                stat = os.stat(DATA_FOLDER + file)
                file_time = datetime.datetime.fromtimestamp(stat.st_mtime)
                if (datetime.datetime.now() - file_time).days > 10:
                    try:
                        os.remove(DATA_FOLDER + file)
                    except Exception as e:
                        pass

    def cache_if_not_busy(self):
        should_cache = False
        with self.background_mutex:
            if len(self.worker_threads) == 0 and self.current_state != None:
                should_cache = True
        if should_cache:
            self.add_background_job(self.current_state[0], self.current_state[1]+1, False)


    def stop(self):
        mixer.music.stop()
        self.current_state = None
    
    def remove_old_threads(self):
        remaining_threads = []
        with self.background_mutex:
            for thread_path, thread_page, thread_fast, thread in self.worker_threads:
                if thread.is_alive():
                    remaining_threads.append((thread_path, thread_page, thread_fast, thread))

            self.worker_threads = remaining_threads

    def run(self):
        while True:
            self.remove_old_threads()
            if self.should_follow_lines and mixer.music.get_pos() == -1 and self.current_state != None:
                self.goto_line_with_text(self.current_state[0], self.current_state[1]+1, None)
            elif self.should_follow_lines and mixer.music.get_pos() != -1:
                self.goto_line_if_changed()

            manager.clean_old_files()
            self.cache_if_not_busy()
            time.sleep(0.1)


app = Flask(__name__)
should_stop = False

manager = Manager()

@app.route("/")
def hello_world():
    return "hello world"

@app.route("/stop")
def stop():
    global manager
    manager.stop()
    return "stopping"

@app.route("/follow")
def follow():
    global manager
    manager.enable_line_follow()
    return "following"

@app.route("/unfollow")
def unfollow():
    global manager
    manager.disable_line_follow()
    return "unfollowing"

@app.route('/read', methods=['POST'])
def read():
    global manager
    content = request.json
    print(content)
    manager.goto_line_with_text(content['path'], int(content['page_number']), content['line_text'])
    return "read"

if __name__ == '__main__':
    thread = threading.Thread(target=manager.run)
    thread.start()
    app.run()



'''
script used to generate sioyek.py file
'''
sioyek_commands = {
    "goto_begining": [False,	False,	False,	True],
    "goto_end": [False,	False,	False,	True],
    "goto_definition": [False,	False,	False,	False],
    "overview_definition": [False,	False,	False,	False],
    "portal_to_definition": [False,	False,	False,	False],
    "next_item": [False,False,	False,	True],
    "previous_item": [ False, False , False, True],
    "set_mark": [ False, True , False, False],
    "goto_mark": [ False, True , False, False],
    "goto_page_with_page_number": [ True, False , False, True],
    "search": [ True, False , False, False],
    "ranged_search": [ True, False , False, False],
    "chapter_search": [ True, False , False, False],
    "move_down": [ False, False , False, False],
    "move_up": [ False, False , False, False],
    "move_left": [ False, False , False, False],
    "move_right": [ False, False , False, False],
    "zoom_in": [ False, False , False, False],
    "zoom_out": [ False, False , False, False],
    "fit_to_page_width": [ False, False , False, False],
    "fit_to_page_height": [ False, False , False, False],
    "fit_to_page_height_smart": [ False, False , False, False],
    "fit_to_page_width_smart": [ False, False , False, False],
    "next_page": [ False, False , False, False],
    "previous_page": [ False, False , False, False],
    "open_document": [ False, False , True, True],
    "debug": [ False, False , False, False],
    "add_bookmark": [ True, False , False, False],
    "add_highlight": [ False, True , False, False],
    "goto_toc": [ False, False , False, False],
    "goto_highlight": [ False, False , False, False],
    "goto_bookmark": [ False, False , False, False],
    "goto_bookmark_g": [ False, False , False, False],
    "goto_highlight_g": [ False, False , False, False],
    "goto_highlight_ranged": [ False, False , False, False],
    "link": [ False, False , False, False],
    "portal": [ False, False , False, False],
    "next_state": [ False, False , False, False],
    "prev_state": [ False, False , False, False],
    "pop_state": [ False, False , False, False],
    "test_command": [ False, False , False, False],
    "delete_link": [ False, False , False, False],
    "delete_portal": [ False, False , False, False],
    "delete_bookmark": [ False, False , False, False],
    "delete_highlight": [ False, False , False, False],
    "goto_link": [ False, False , False, False],
    "goto_portal": [ False, False , False, False],
    "edit_link": [ False, False , False, False],
    "edit_portal": [ False, False , False, False],
    "open_prev_doc": [ False, False , False, False],
    "open_document_embedded": [ False, False , False, False],
    "open_document_embedded_from_current_path": [ False, False , False, False],
    "copy": [ False, False , False, False],
    "toggle_fullscreen": [ False, False , False, False],
    "toggle_one_window": [ False, False , False, False],
    "toggle_highlight": [ False, False , False, False],
    "toggle_synctex": [ False, False , False, False],
    "command": [ False, False , False, False],
    "external_search": [ False, True , False, False],
    "open_selected_url": [ False, False , False, False],
    "screen_down": [ False, False , False, False],
    "screen_up": [ False, False , False, False],
    "next_chapter": [ False, False , False, True],
    "prev_chapter": [ False, False , False, True],
    "toggle_dark_mode": [ False, False , False, False],
    "toggle_presentation_mode": [ False, False , False, False],
    "toggle_mouse_drag_mode": [ False, False , False, False],
    "close_window": [ False, False , False, False],
    "quit": [ False, False , False, False],
    "q": [ False, False , False, False],
    "open_link": [ True, False , False, False],
    "keyboard_select": [ True, False , False, False],
    "keyboard_smart_jump": [ True, False , False, False],
    "keyboard_overview": [ True, False , False, False],
    "keys": [ False, False , False, False],
    "keys_user": [ False, False , False, False],
    "prefs": [ False, False , False, False],
    "prefs_user": [ False, False , False, False],
    "import": [ False, False , False, False],
    "export": [ False, False , False, False],
    "enter_visual_mark_mode": [ False, False , False, False],
    "move_visual_mark_down": [ False, False , False, False],
    "move_visual_mark_up": [ False, False , False, False],
    "set_page_offset": [ True, False , False, False],
    "toggle_visual_scroll": [ False, False , False, False],
    "toggle_horizontal_scroll_lock": [ False, False , False, False],
    "toggle_custom_color": [ False, False , False, False],
    "execute": [ True, False , False, False],
    "execute_predefined_command": [ False, True, False, False],
    "embed_annotations": [ False, False, False, False],
    "copy_window_size_config": [ False, False, False, False],
    "toggle_select_highlight": [ False, False, False, False],
    "set_select_highlight_type": [ False, True, False, False],
    "open_last_document": [ False, False, False, False],
    "toggle_window_configuration": [ False, False, False, False],
    "prefs_user_all": [ False, False, False, False],
    "keys_user_all": [ False, False, False, False],
    "fit_to_page_width_ratio": [ False, False, False, False],
    "smart_jump_under_cursor": [ False, False, False, False],
    "overview_under_cursor": [ False, False, False, False],
    "close_overview": [ False, False, False, False],
    "visual_mark_under_cursor": [ False, False, False, False],
    "close_visual_mark": [ False, False, False, False],
    "zoom_in_cursor": [ False, False, False, False],
    "zoom_out_cursor": [ False, False, False, False],
    "goto_left": [ False, False, False, False],
    "goto_left_smart": [ False, False, False, False],
    "goto_right": [ False, False, False, False],
    "goto_right_smart": [ False, False, False, False],
    "rotate_clockwise": [ False, False, False, False],
    "rotate_counterclockwise": [ False, False, False, False],
    "goto_next_highlight": [ False, False, False, False],
    "goto_prev_highlight": [ False, False, False, False],
    "goto_next_highlight_of_type": [ False, False, False, False],
    "goto_prev_highlight_of_type": [ False, False, False, False],
    "add_highlight_with_current_type": [ False, False, False, False],
    "enter_password": [ True, False , False, False],
    "toggle_fastread": [ False, False , False, False],
    "goto_top_of_page": [ False, False , False, False],
    "goto_bottom_of_page": [ False, False , False, False],
    "new_window": [ False, False , False, False],
    "toggle_statusbar": [ False, False , False, False],
    "reload": [ False, False , False, False],
    "synctex_under_cursor": [ False, False , False, False],
    "set_status_string": [ True, False , False, False],
    "clear_status_string": [ False, False , False, False],
    "toggle_titlebar": [ False, False , False, False],
    "next_preview": [ False, False , False, False],
    "previous_preview": [ False, False , False, False],
    "goto_overview": [ False, False , False, False],
    "portal_to_overview": [ False, False , False, False],
    }

def print_method(command_name, requires_text, requires_symbol, requires_filename):
    res = ''
    res += 'def {}(self'.format(command_name)
    if requires_text:
        res += ', text'
    elif requires_symbol:
        res += ', symbol'
    elif requires_filename:
        res += ', filename'
    res += ', focus=False):\n'
    if requires_text:
        res += ' ' * 4 + 'data = text\n'
    elif requires_symbol:
        res += ' ' * 4 + 'data = symbol\n'
    elif requires_filename:
        res += ' ' * 4 + 'data = filename\n'
    else:
        res += ' ' * 4 + 'data = None\n'

    
    res += ' ' * 4 + 'self.run_command("{}", data, focus=focus)\n'.format(command_name)
    print(res)

if __name__ == '__main__':
    for command_name, (requires_text, requires_symbol, requires_filename, _) in sioyek_commands.items():
        print_method(command_name, requires_text, requires_symbol, requires_filename)



from flask import Flask, request, jsonify
from tqdm import tqdm as q
from transformers import ReformerModelWithLMHead
import re
import torch

def encode(list_of_strings, pad_token_id=0):
    max_length = max([len(string) for string in list_of_strings])

    # create emtpy tensors
    attention_masks = torch.zeros((len(list_of_strings), max_length), dtype=torch.long)
    input_ids = torch.full((len(list_of_strings), max_length), pad_token_id, dtype=torch.long)

    for idx, string in enumerate(list_of_strings):
        # make sure string is in byte format
        if not isinstance(string, bytes):
            string = str.encode(string)

        input_ids[idx, :len(string)] = torch.tensor([x + 2 for x in string])
        attention_masks[idx, :len(string)] = 1

    return input_ids, attention_masks
    
# Decoding
def decode(outputs_ids):
    decoded_outputs = []
    for output_ids in outputs_ids.tolist():
        # transform id back to char IDs < 2 are simply transformed to ""
        decoded_outputs.append("".join([chr(x - 2) if x > 1 else "" for x in output_ids]))
    return decoded_outputs

def roll_encoding(encoding):
    # by default if encodings are smaller than maximum length, they are padded with
    # zeroes at the end. We need the zero padding to be at the begining for next
    # character prediction to work
    new_encoding = torch.clone(encoding)
    num_rows = encoding.shape[0]
    for i in range(num_rows):
        roll_amount = int(len(encoding[i]) - encoding[i].nonzero().max()-1)
        new_encoding[i] = encoding[i].roll(roll_amount)
    return new_encoding

def get_page_queries(page_text, context_size=49):
    res = []
    for i in q(range(1, len(page_text))):
        context = page_text[max(i-context_size, 0): i]
        label = page_text[i]
        res.append((clean_text(context), label))
    return res

def importance_mask(predicted, query):
    # we define important characters to be the ones which are predicted incorrectly
    # given their context
    mask = []
    for pred, q in zip(predicted, query):
        target = q[1]
        if pred[-1] == target:
            mask.append(0)
        else:
            mask.append(1)
    return mask

def batchify(seq, size):
    # we have to batch the data to prevent OOM erros
    res = []
    while seq != []:
        res.append(seq[:size])
        seq = seq[size:]
    return res

def batch_predictor_cuda(texts, batch_size=250):
    
    res = []
    for batch in q(batchify(texts, batch_size)):
        # predict only one character ahead
        max_length = max([len(x) for x in texts]) + 1
        encoded, _ = encode(batch)
        encoded = roll_encoding(encoded)
        generated = model.generate(encoded.to(device), max_length=max_length)
        decoded = decode(generated)
        del encoded
        del generated
        res.extend(decoded)
    return res

def split_text_and_imask(text, imask):
    # it's like text.split(( |\n)), except we split the mask at the same locations too
    text_splits = []
    imask_splits = []
    
    split_locations = [x.start() for x in re.finditer('( |\n)', text)]
    split_locations += [len(text)]
    
    if len(split_locations) > 0:
        text_splits.append(text[:split_locations[0]])
        imask_splits.append(imask[:split_locations[0]])
        for i in range(1, len(split_locations)):
            text_splits.append(text[split_locations[i-1]+1:split_locations[i]])
            imask_splits.append(imask[split_locations[i-1]+1:split_locations[i]])
    else:
        text_splits = [text]
    
    return text_splits, imask_splits

def clean_text(text):
    return "".join(c for c in text if ord(c) < 127)

def refine_imasks(imask_list, fill=True):
    # convert all zeroes before the last one to ones. For example:
    # [1,0,0,1,0,0,0,0] => [1,1,1,1,0,0,0,0]
    # if `fill==True and refine==True`: if there are too much ones, highlight the whole character
    # instead of highlighting most of it
    res = []
    for imask in imask_list:
        try:
            last_index = len(imask) - imask[::-1].index(1)
            if fill and (last_index > len(imask) / 2):
                res.append([1 for x in range(len(imask))])
            else:
                res.append([1 for _ in range(last_index)] + [0 for _ in range(len(imask) - last_index)])
        except ValueError as e:
            # this happens when the mask has no 1s
            res.append(imask)
    return res

def create_result_string(refined_imasks):
    res = ""
    for mask in refined_imasks:
        res = res + "".join([str(x) for x in mask]) + "0"
    return res[:-1]

def get_mask_from_text(text, refine=True, fill=True, context_size=49):
    queries = get_page_queries(text, context_size)
    query_texts = [q[0] for q in queries]
    predicted_results = batch_predictor_cuda(query_texts)
    imask = [1] + importance_mask(predicted_results, queries)
    
    text_splits, imask_splits = split_text_and_imask(text, imask)
    if refine:
        refined_imasks = refine_imasks(imask_splits, fill=fill)
    else:
        refined_imasks = imask_splits
    return create_result_string(refined_imasks)

if __name__ == '__main__':
    device = torch.device('cuda')
    model = ReformerModelWithLMHead.from_pretrained("google/reformer-enwik8").to(device)

    app = Flask(__name__)

    @app.route('/', methods=['POST', 'GET'])
    def handle_text_mask():
        text = request.values.get('text')
        page = request.values.get('page')
        should_refine = int(request.values.get('refine'))
        should_fill = int(request.values.get('fill'))
        context_size = int(request.values.get('context_size'))
        
        refined_imask =  get_mask_from_text(f'{text}', should_refine, should_fill, context_size)
        print('-' * 80)
        print('len(text):', len(text))
        print('len(imask):', len(refined_imask))
            
        return jsonify({'text': refined_imask, 'page': page})

    app.run()


'''
This script was tested with PyMuPDF version 1.17.6. Other versions may require slight modification of the code.

This script can be used to embed annotations as you create them, so they are viewable in other
PDF viewers.
This is basically a script that can either add a bookmark or a highlight to a page based on the command line arguments.
Here is the `prefs_user.config` that I used:

	execute_command_b python path/to/embedded_annotations.py bookmark "%1" %{mouse_pos_document} "%5"
	execute_command_h python path/to/embedded_annotations.py highlight "%1" "%4" "%3"

And here is a `keys_user.config` file that can be used:

	execute_command_b b
	execute_command_h;add_highlight h

which basically adds both sioyek and embedded highlights and bookmarks (if `ADD_BOOKMARKS_TO_SIOYEK` is True). Alternatively, you can use
a different keybinding for embedded annotations so you have control over what happens. Now you can use alt+b or alt+h to bookmark/highlight.

	execute_command_b <A-b>
	execute_command_h <A-h>
'''

import sys
import fitz
import subprocess

# if set to true, we re-add the bookmarks into sioyek, so we have both types of bookmarks
ADD_BOOKMARKS_TO_SIOYEK = False
PATH_TO_SIOYEK = r'path/to/sioyek.exe'

def add_bookmark(doc_path, page_number, location, text):
	doc = fitz.open(doc_path)
	page = doc.loadPage(page_number)
	page.addTextAnnot(location, text)
	doc.saveIncr()
	doc.close()
	if ADD_BOOKMARKS_TO_SIOYEK:
		subprocess.run([PATH_TO_SIOYEK, '--execute-command', 'add_bookmark','--execute-command-data', text])
		subprocess.run([PATH_TO_SIOYEK, '--execute-command', 'reload'])


def add_highlight(doc_path, page_number, text):
	doc = fitz.open(doc_path)
	page = doc.loadPage(page_number)
	quads = page.searchFor(text, flags=fitz.TEXT_PRESERVE_WHITESPACE, hit_max=50)
	page.addHighlightAnnot(quads)
	doc.saveIncr()
	doc.close()

if __name__ == '__main__':
	if sys.argv[1] == 'bookmark':
		add_bookmark(sys.argv[2], int(sys.argv[3]), (float(sys.argv[4]), float(sys.argv[5])), sys.argv[6])

	if sys.argv[1] == 'highlight':
		add_highlight(sys.argv[2], int(sys.argv[3]), sys.argv[4])

